#ifndef __BITVECTOR_HPP__
#define __BITVECTOR_HPP__

#include <stdio.h>
#include <cstring>
#include "common.hpp"
#include "verticalCode.hpp"
using namespace std;

#define BLOCKSIZE 32

class bitVector{
public:
	bitVector(uint _size = 0): size(_size), blockSize((_size + BLOCKSIZE - 1)/BLOCKSIZE), B(NULL){
		if (size){
			B = new uint [blockSize];
			if (B == NULL){
				printf("cannot allocate for B\n");
				throw;
			}
			memset(B,0,sizeof(uint)*blockSize);
		}
	}

	~bitVector(){
		if (B) delete[] B;
	}

	void resize(uint n){
		uint newBlockSize = (n+BLOCKSIZE-1)/BLOCKSIZE;
		uint* newB = new uint [newBlockSize];
		if (newB == NULL){
			printf("cannot allocate at bitVector::resize()");
			throw;
		}
		memset(newB,0,sizeof(uint)*newBlockSize);
		uint copySize = min(newBlockSize,blockSize);
		for (uint i = 0; i < copySize; i++){
			newB[i] = B[i];
		}
		if (B) delete[] B;
		B = newB;
		size = n;
		blockSize = newBlockSize;
	}

	void setBit(uint pos, uint x){
		uint blockPos = pos / BLOCKSIZE;
		uint offset   = pos % BLOCKSIZE;
		if      (x == 0) B[blockPos] &= (~(1U << offset));
		else if (x == 1) B[blockPos] |=   (1U << offset);
		else {
			fprintf(stderr,"unexpected value in setBit. x = %d\n",x);
			throw;
		}
	}

	void setBits(uint pos, uint width, uint x){
		for (uint i = 0; i < width; i++){
			setBit(i+pos,(x >> i)&1);
		}
	}

	uint getBit(uint pos)const {
		uint blockPos = pos / BLOCKSIZE;
		uint offset   = pos % BLOCKSIZE;

		if (pos >= size){
			fprintf(stderr,"error in bitVector::getBit() pos:%d size:%d\n",pos,size);
			exit(1);
		}
		return (B[blockPos] >> offset) & 1;
	}

	uint getBits(uint pos, uint width) const{
		if (width >= 32){
			fprintf(stderr,"getBits error. width:%d\n",width);
			exit(1);
		}
		if (pos+width > size){
			fprintf(stderr,"getBits error. pos:%d width:%d size:%d\n",pos,width,size);
			exit(1);
		}
		if (width == 0) return 0;

		/*
		uint x = 0;
		for (uint i = 0; i < width; i++){
			x += getBit(pos+i) << i;
		}
		*/
		
		//block version
		const uint endPos = pos+width-1;
		const uint blockPos_1 = pos / BLOCKSIZE;
		const uint blockPos_2 = endPos / BLOCKSIZE;
		const uint offset_1   = pos % BLOCKSIZE;
	
		if (blockPos_1 == blockPos_2){
			return (B[blockPos_1] >> offset_1) & ((1U << width) - 1);
		}
		else {
			const uint offset_2   = endPos % BLOCKSIZE;
			return  (B[blockPos_1] >> offset_1) + ((B[blockPos_2] & ((1U << (offset_2+1))-1)) << (32-offset_1));
		}		

	}



	uint* getB() const { return B; }

	uint getBlock(uint i) const { return B[i]; }

	uint getSize() const{
		return size;
	}

	uint getBlockSize() const{
		return blockSize;
	}

	void printBits(){
		for (uint i = 0; i < size; i++){
			if (getBit(i)) putchar('1');
			else           putchar('0');
			if ((i+1) % 8 == 0) putchar(' ');
			if ((i+1) % 32 == 0) putchar('\n');
		}
	}

	void copy(bitVector& from, uint bitPos, uint bitNum){
		for (uint i = 0; i < bitNum; i++){
			uint x = from.getBit(i);
			setBit(bitPos+i,x);
		}
	}

	void write(FILE* outfp){
		fwrite(&blockSize,sizeof(uint),1,outfp);
		fwrite(B,sizeof(uint),blockSize,outfp);
	}

	void read(FILE* infp){
		uint _blockSize = 0;
		fread(&_blockSize,sizeof(uint),1,infp);
		resize(_blockSize * 32);
		fread(B,sizeof(uint),blockSize,infp);
		blockSize = _blockSize;
	}

//---: use rank -----

#define LEVELA (1 << 8)
#define LEVELB (1 << 5)
#define B_LEVELA (LEVELA/BLOCKSIZE) //32
#define B_LEVELB (LEVELB/BLOCKSIZE) //4

	uint _popCount(const uint x) const{
		return popCount[ x >> 24        ] + 
			   popCount[(x >> 16) & 0xFF] + 
			   popCount[(x >> 8 ) & 0xFF] + 
			   popCount[x & 0xFF];	
	}

	void build(){
		uint levelASize = size / LEVELA + 1;
		uint levelBSize = size / LEVELB + 1;

		levelA = new uint[levelASize];
		levelB = new uchar[levelBSize];

		for (uint i = 0; i < levelASize; i++){
			levelA[i] = 0;
		}
		for (uint i = 0; i < levelBSize; i++){
			levelB[i] = 0;
		}

		uint r = 0;
		for (uint ia = 0; ia <= size; ia += LEVELA){
			levelA[ia/LEVELA] = r;
			for (uint ib = 0; ib < LEVELA && (ia+ib) <= size; ib += LEVELB){
				levelB[(ia+ib)/LEVELB] = r - levelA[ia/LEVELA];
				r += _popCount(B[(ia+ib)/LEVELB]);
			}
		}
	}

	uint rank(uint pos) const{
		const uint remain    = pos % BLOCKSIZE;
		const uint x = B[pos/BLOCKSIZE] & ((1 << remain) - 1);
		
		return levelA[pos/LEVELA] + 
			   levelB[pos/LEVELB] + 
		       _popCount(x);
	}

public:
	uint getAllocSize() const{
		return size + (sizeof(uint)*(size / LEVELA + 1) + sizeof(uchar)*(size / LEVELB + 1))*8;
	}
	
private:
	uint* B;
	uint size;
	uint blockSize;
	
	uint* levelA;
	uchar* levelB;
};

#endif
