#ifndef __SSV_HPP__
#define __SSV_HPP__

#include "common.hpp"

#define SSV_BLOCK_SHIFT  (5)
#define SSV_BLOCK        (1U << SSV_BLOCK_SHIFT)

#define SSV_LBLOCK_SHIFT (8)
#define SSV_LBLOCK		 (1U << SSV_LBLOCK_SHIFT)
#define SSV_MBLOCK_SHIFT (6)
#define SSV_MBLOCK       (1U << SSV_MBLOCK_SHIFT)

class ssv{

public:
	ssv():size(0),B(NULL),levelL(NULL),levelM(NULL){}
	ssv(const uint _size):B(NULL),levelL(NULL),levelM(NULL){
		resize(_size);
	}

	void resize(const uint _size){
		size = _size;
		blockSize = (size+SSV_BLOCK-1)/SSV_BLOCK;
		oneNum = 0;
		zeroNum = 0;
		isBuild = false;
		B = new uint [blockSize];
		if (B == NULL){
			throw "cannot allocate for ssv::B";
		}
		memset(B,0,sizeof(uint)*blockSize);

		LBlockSize = size / SSV_LBLOCK + 1;
		MBlockSize = size / SSV_MBLOCK + 1;

		levelL = new uint [LBlockSize];
		levelM = new uchar [MBlockSize];
	}

	~ssv(){
		if (B) delete[] B;
		if (levelL) delete[] levelL;
		if (levelM) delete[] levelM;
	}

	uint getBit(const uint pos) const{
		return (B[pos / SSV_BLOCK] >> (pos % SSV_BLOCK)) & 1;
	}

	uint rank(uint pos, const uint bit) const{
		pos++;
		if (bit == 0) return pos-_rank1(pos);
		else		  return _rank1(pos);
	}

	uint select(const uint pos, const uint bit) const{
		if (bit) return _select1(pos);
		else     return _select0(pos);
	}

	void setBit(uint pos, uint x){
		if (!x) B[pos / SSV_BLOCK] &= (~(1U << (pos % SSV_BLOCK)));
		else    B[pos / SSV_BLOCK] |=   (1U << (pos % SSV_BLOCK));
	}

	uint getAllocSize() const{
		uint sum = size;
		if (isBuild){
			sum += LBlockSize * 32 + MBlockSize * 8;
		}
		return sum;
	}

	void build(){
		uint sum = 0;
		for (uint il = 0; il <= size ;il += SSV_LBLOCK){
			levelL[il/SSV_LBLOCK] = sum;
			for (uint im = 0; im < SSV_LBLOCK && il+im <= size; im += SSV_MBLOCK){
				levelM[(il+im)/SSV_MBLOCK] = sum - levelL[il/SSV_LBLOCK];
				for (uint is = 0; is < SSV_MBLOCK && (il+im+is) <= size; is += SSV_BLOCK){
					sum += popCount(B[(il+im+is)/SSV_BLOCK]);
				}
				
			}
			
		}
		isBuild = true;
	}

private:
	uint _rank1(const uint pos) const{
		const uint prePopCount= ((pos / SSV_BLOCK) & 1) ? popCount(B[pos/SSV_BLOCK-1]) : 0;
		return levelL[pos >> SSV_LBLOCK_SHIFT] + 
			   levelM[pos >> SSV_MBLOCK_SHIFT] + 
			   prePopCount + 
			   popCount((B[pos >> SSV_BLOCK_SHIFT] & ((1U << (pos  % SSV_BLOCK))-1)));
	}

	uint _select0(uint x) const{
		uint left  = 0;
		uint right = LBlockSize;
		while (left < right){
			uint mid = (left+right)/2;
			if (SSV_LBLOCK * mid - levelL[mid] < x) left  = mid+1;
			else                 right = mid;
		}
		uint posL = (left != 0) ? left-1: 0;
		
		 //sequential search over levelB
		uint posM = posL * (SSV_LBLOCK/SSV_MBLOCK);
		x -= (SSV_LBLOCK*posL - levelL[posL]);

		uint mstep = SSV_LBLOCK/SSV_MBLOCK;
		while ((posM+1 < MBlockSize) && (((posM+1)%mstep) != 0) && (SSV_MBLOCK*((posM+1)%mstep)-levelM[posM+1] < x)){
			posM++;
		}

		x -= (SSV_MBLOCK*(posM % mstep) - levelM[posM]);

		uint ret    = posM * SSV_MBLOCK;
		uint posB   = posM * (SSV_MBLOCK/SSV_BLOCK);
		uint rank32 = 32- popCount(B[posB]);
		if (rank32 < x){
			posB++;
			ret += SSV_BLOCK;
			x -= rank32;
		}

		uint curB   = B[posB];
		uint rank16 = 16-popCount(curB & ((1U<<16)-1));
		if (rank16 < x){
			curB >>= 16;
			ret += 16;
			x -= rank16;
		}

		uint rank8  = 8-popCount(curB & ((1U << 8)-1));
		if (rank8 < x){
			curB >>= 8;
			ret += 8;
			x -= rank8;
		}
		
		while (x > 0){
			if ((~curB) & 1){
				x--;
			}
			curB >>= 1;
			ret++;
		}
		return ret-1;
	}

	uint _select1(uint x) const{
		uint left  = 0;
		uint right = LBlockSize;
		while (left < right){
			uint mid = (left+right)/2;
			if (levelL[mid] < x) left  = mid+1;
			else                 right = mid;
		}
		uint posL = (left != 0) ? left-1: 0;
		
		 //sequential search over levelB
		uint posM = posL * (SSV_LBLOCK/SSV_MBLOCK);
		x -= levelL[posL];
		while ((posM+1 < MBlockSize) && (((posM+1)%(SSV_LBLOCK/SSV_MBLOCK)) != 0) && (levelM[posM+1] < x)){
			posM++;
		}

		x -= levelM[posM];

		uint ret    = posM * SSV_MBLOCK;
		uint posB   = posM * (SSV_MBLOCK/SSV_BLOCK);
		uint rank32 = popCount(B[posB]);
		if (rank32 < x){
			posB++;
			ret += SSV_BLOCK;
			x -= rank32;
		}

		uint curB   = B[posB];
		uint rank16 = popCount(curB & ((1U<<16)-1));
		if (rank16 < x){
			curB >>= 16;
			ret += 16;
			x -= rank16;
		}

		uint rank8  = popCount(curB & ((1U << 8)-1));
		if (rank8 < x){
			curB >>= 8;
			ret += 8;
			x -= rank8;
		}
		
		while (x > 0){
			if (curB & 1){
				x--;
			}
			curB >>= 1;
			ret++;
		}
		return ret-1;
	}


	uint* B;
	uint size;
	uint blockSize;	// (size+SSV_BLOCK-1)/SSV_BLOCKSZIE
	uint oneNum;
	uint zeroNum;

	uint LBlockSize;
	uint MBlockSize;

	uint* levelL;
	uchar* levelM;

	bool isBuild;
};


#endif
